# -*- coding: utf-8 -*-

import copy
import json
import os
import shutil
import subprocess
import tarfile
import urllib.parse

from . import config, repo_local, repo_online, repo_publish
from .container import Container
from .image import Image

DEFINITION_MEMBERS = {'version', 'meta', 'autostart', 'containers'}

class App:
    def __init__(self, name, define_containers=True, load_from_repo=True):
        self.name = name
        self.version = None
        self.app_dir = os.path.join(config.APPS_DIR, name)
        self.meta = {}
        self.autostart = False
        self.containers = []
        if load_from_repo:
            self.set_definition(repo_local.get_app(name), define_containers)

    def set_definition(self, definition, define_containers):
        # Set attributes given by definition
        for key in DEFINITION_MEMBERS.intersection(definition):
            setattr(self, key, definition[key])
        # Populate containers property with actual container objects
        self.containers = [Container(container, define_containers) for container in definition['containers']]

    def get_definition(self):
        # Return shallow copy of image definition as dictionary
        definition = {}
        for key in DEFINITION_MEMBERS:
            value = getattr(self, key)
            if value:
                definition[key] = copy.copy(value)
        # Overwrite containers key with list of container names
        definition['containers'] = [container.name for container in self.containers]
        return definition

    def download(self, observer=None):
        # Download the archive with application scripts and install data
        os.makedirs(config.TMP_APPS_DIR, 0o700, True)
        archive_url = urllib.parse.urljoin(config.ONLINE_APPS_URL, f'{self.name}.tar.xz')
        archive_path = os.path.join(config.TMP_APPS_DIR, f'{self.name}.tar.xz')
        definition = repo_online.get_app(self.name)
        if observer:
            observer.units_total = definition['dlsize']
        repo_online.download_archive(archive_url, archive_path, definition['hash'], observer)

    def unpack_downloaded(self, observer=None):
        # Unpack downloaded archive with application scripts and install data
        archive_path = os.path.join(config.TMP_APPS_DIR, f'{self.name}.tar.xz')
        definition = repo_online.get_app(self.name)
        if observer:
            observer.units_total = definition['size']
        repo_online.unpack_archive(archive_path, config.APPS_DIR, definition['hash'], observer)

    def run_script(self, action):
        # Runs script for an app, if the script is present
        script_dir = os.path.join(self.app_dir, action)
        script_path = os.path.join(self.app_dir, f'{script_dir}.sh')
        if os.path.exists(script_path):
            # Run the script in its working directory, if there is one, so it doesn't have to figure out paths to packaged files
            env = os.environ.copy()
            env['LAYERS_DIR'] = config.LAYERS_DIR
            env['VOLUMES_DIR'] = config.VOLUMES_DIR
            env['APPS_DIR'] = config.APPS_DIR
            env['LOG_DIR'] = config.LOG_DIR
            cwd = script_dir if os.path.exists(script_dir) else self.app_dir
            subprocess.run(script_path, cwd=cwd, env=env, check=True)

    def create_container(self, name, definition):
        # Create container and enhance its definition (typically mounts) based on application requirements
        container = Container(name, False)
        container.set_definition(Image(definition['image']).get_definition())
        if 'depends' in definition:
            container.depends = definition['depends']
        if 'env' in definition:
            container.env.update(definition['env'])
        if 'mounts' in definition:
            container.mounts.update(definition['mounts'])
        container.create()
        self.containers.append(container)

    def install(self, observer=None):
        # Install the application
        definition = repo_online.get_app(self.name)
        self.version = definition['version']
        self.meta = definition['meta']
        self.run_script('uninstall')
        # Build containers
        for container,container_defintion in definition['containers'].items():
            self.create_container(container, container_defintion)
        # Run install script and register the app
        self.run_script('install')
        repo_local.register_app(self.name, self.get_definition())

    def update(self, observer=None):
        # Remove containers
        for container in self.containers:
            container.destroy()
        # Load online definition
        definition = repo_online.get_app(self.name)
        self.version = definition['version']
        self.meta = definition['meta']
        # Build containers
        for container,container_defintion in definition['containers'].items():
            self.create_container(container, container_defintion)
        # Run update script and re-register the app
        self.run_script('update')
        repo_local.register_app(self.name, self.get_definition())

    def uninstall(self, observer=None):
        # Make sure the containers are stopped
        for container in self.containers:
            container.stop()
        # Run uninstall script
        self.run_script('uninstall')
        # Remove containers
        for container in self.containers:
            container.destroy()
        # Unregister app and remove scripts
        repo_local.unregister_app(self.name)
        try:
            shutil.rmtree(self.app_dir)
        except FileNotFoundError:
            pass

    def start(self, observer=None):
        # Start all application containers
        if observer:
            observer.units_total = len(self.containers)
        for container in self.containers:
            container.start()
            if observer:
                observer.units_done += 1

    def stop(self, observer=None):
        # Stop all application containers
        if observer:
            observer.units_total = len(self.containers)
        for container in self.containers:
            container.stop()
            if observer:
                observer.units_done += 1

    def status(self):
        # Return status fo all application containers
        return {container.name:container.get_state() for container in sorted(self.containers)}

    def set_autostart(self, autostart):
        # Configure if the application should be automatically started after boot
        self.autostart = autostart
        repo_local.register_app(self.name, self.get_definition())

    def publish(self, filename):
        # Create application archive and register to publish repository
        builddir = os.path.dirname(filename)
        os.makedirs(config.PUB_APPS_DIR, 0o755, True)
        files = repo_publish.TarSizeCounter()
        archive_path = os.path.join(config.PUB_APPS_DIR, f'{self.name}.tar.xz')
        with tarfile.open(archive_path, 'w:xz') as tar:
            for content in ('install', 'install.sh', 'update', 'update.sh', 'uninstall', 'uninstall.sh'):
                content_path = os.path.join(builddir, content)
                if os.path.exists(content_path):
                    tar.add(content_path, os.path.join(self.name, content), filter=files.add_file)
        with open(filename) as f:
            definition = json.load(f)
        definition['size'] = files.size
        definition['dlsize'] = os.path.getsize(archive_path)
        definition['hash'] = repo_publish.sign_file(archive_path).hex()
        repo_publish.register_app(self.name, definition)
        return (definition['size'], definition['dlsize'])

    def unpublish(self):
        # Remove the application from publish repository
        repo_publish.unregister_app(self.name)
        archive_path = os.path.join(config.PUB_APPS_DIR, f'{self.name}.tar.xz')
        try:
            os.unlink(archive_path)
        except FileNotFoundError:
            pass
