# -*- coding: utf-8 -*-

import json
import os
import requests
import shutil
import tarfile
import time
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, utils
from cryptography.hazmat.primitives.serialization import load_pem_public_key

from . import config
from .exceptions import AppNotFoundError, ImageNotFoundError

TYPE_APP = 'apps'
TYPE_IMAGE = 'images'

public_key = None

def get_public_key():
    global public_key
    if not public_key:
        pem = f'-----BEGIN PUBLIC KEY-----\n{config.ONLINE_PUBKEY}\n-----END PUBLIC KEY-----'
        public_key = load_pem_public_key(pem.encode(), default_backend())
    return public_key

def verify_fileobj(fileobj, expected_hash):
    hasher = hashes.Hash(hashes.SHA512(), default_backend())
    while True:
        data = fileobj.read(64*1024)
        if not data:
            break
        hasher.update(data)
    get_public_key().verify(bytes.fromhex(expected_hash), hasher.finalize(), ec.ECDSA(utils.Prehashed(hashes.SHA512())))

def download_archive(archive_url, archive_path, expected_hash, observer=None):
    # Check if an archive needs to be downloaded via http(s)
    do_download = True
    # If the file already exists in the temporary directory, verify the signature
    if os.path.exists(archive_path):
        try:
            with open(archive_path, 'rb') as f:
                verify_fileobj(f, expected_hash)
            # If the signature matches, skip download
            if observer:
                observer.units_done = os.path.getsize(archive_path)
            do_download = False
        except InvalidSignature:
            # If the signature is invalid, redownload the file
            pass
    if do_download:
        # Download archive via http(s) and store in temporary directory
        with open(archive_path, 'wb') as f, requests.Session() as session:
            resource = session.get(archive_url, stream=True)
            resource.raise_for_status()
            if observer:
                for chunk in resource.iter_content(chunk_size=64*1024):
                    if chunk:
                        observer.units_done += f.write(chunk)
            else:
                for chunk in resource.iter_content(chunk_size=64*1024):
                    if chunk:
                        f.write(chunk)

def unpack_archive(archive_path, destination, expected_hash, observer):
    with open(archive_path, 'rb') as f:
        # Verify file object, then seek back and open it as tar without losing handle, preventing possible malicious race conditions
        verify_fileobj(f, expected_hash)
        f.seek(0)
        # Remove the target directory, if it exists from previous failed installation
        dst_dir = os.path.join(destination, os.path.basename(archive_path)[:-7])
        try:
            shutil.rmtree(dst_dir)
        except FileNotFoundError:
            pass
        # Extract the tar members while counting their size
        # If this is done as non-root, extractall() from https://github.com/python/cpython/blob/master/Lib/tarfile.py needs to be reimplemented instead
        tar = tarfile.open(fileobj=f)
        if observer:
            for tarinfo in tar:
                tar.extract(tarinfo, destination, numeric_owner=True)
                observer.units_done += tarinfo.size
        else:
            tar.extractall(destination, numeric_owner=True)
    # Remove the archive
    os.unlink(archive_path)

data = None

def load(force=False):
    global data
    if not data or force:
        with requests.Session() as session:
            resource = session.get(config.ONLINE_REPO_URL, timeout=5)
            resource.raise_for_status()
            packages = resource.content
            resource = session.get(config.ONLINE_SIG_URL, timeout=5)
            resource.raise_for_status()
            packages_sig = resource.content
        get_public_key().verify(packages_sig, packages, ec.ECDSA(hashes.SHA512()))
        data = json.loads(packages.decode())

def get_entries(entry_type):
    load()
    return data[entry_type]

def get_entry(entry_type, name, exception):
    try:
        return get_entries(entry_type)[name]
    except KeyError as e:
        raise exception(name) from e

def get_images():
    return get_entries(TYPE_IMAGE)

def get_image(image_name):
    return get_entry(TYPE_IMAGE, image_name, ImageNotFoundError)

def get_apps():
    return get_entries(TYPE_APP)

def get_app(app_name):
    return get_entry(TYPE_APP, app_name, AppNotFoundError)
